import sqlite3
import time


class Client(object):
    def __init__(self, conn_string):
        """
        A SQLite-based ``Client``.

        Args:
            conn_string (str): The DSN. The host/port/db are parsed out of it.
                Should be of the format ``sqlite:///path/to/db/file.db``
        """
        # This is actually the filepath to the DB file.
        self.conn_string = conn_string
        # Kill the 'sqlite://' portion.
        path = self.conn_string.split("://", 1)[1]
        self.conn = sqlite3.connect(path)

    def _run_query(self, query, args):
        cur = self.conn.cursor()

        if not args:
            cur.execute(query)
        else:
            cur.execute(query, args)

        self.conn.commit()
        return cur

    def setup_tables(self, queue_name="all"):
        """
        Allows for manual creation of the needed tables.

        Args:
            queue_name (str): Optional. The name of the queue. Default is
                `all`.
        """
        # For manually creating the tables...
        query = (
            "CREATE TABLE `queue_{}` "
            "(task_id text, data text, delay_until integer)"
        ).format(queue_name)
        self._run_query(query, None)

    def len(self, queue_name):
        """
        Returns the length of the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                `Gator`` instance.

        Returns:
            int: The length of the queue
        """
        query = "SELECT COUNT(task_id) FROM `queue_{}`".format(queue_name)
        cur = self._run_query(query, [])
        res = cur.fetchone()
        return res[0]

    def drop_all(self, queue_name):
        """
        Drops all the task in the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.
        """
        query = "DELETE FROM `queue_{}`".format(queue_name)
        self._run_query(query, [])

    def push(self, queue_name, task_id, data, delay_until=None):
        """
        Pushes a task onto the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.
            task_id (str): The identifier of the task.
            data (str): The relevant data for the task.
            delay_until (float): Optional. The Unix timestamp to delay
                processing of the task until. Default is `None`.

        Returns:
            str: The task ID.
        """
        if delay_until is None:
            delay_until = time.time()

        query = (
            "INSERT INTO `queue_{}` "
            "(task_id, data, delay_until) "
            "VALUES (?, ?, ?)"
        ).format(queue_name)
        self._run_query(query, [task_id, data, int(delay_until)])
        return task_id

    def pop(self, queue_name):
        """
        Pops a task off the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.

        Returns:
            str: The data for the task.
        """
        now = int(time.time())
        query = (
            "SELECT task_id, data, delay_until "
            "FROM `queue_{}` "
            "WHERE delay_until <= ? "
            "LIMIT 1"
        ).format(queue_name)
        cur = self._run_query(query, [now])
        res = cur.fetchone()

        if res:
            query = "DELETE FROM `queue_{}` WHERE task_id = ?".format(
                queue_name
            )
            self._run_query(query, [res[0]])

            return res[1]

    def get(self, queue_name, task_id):
        """
        Pops a specific task off the queue by identifier.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.
            task_id (str): The identifier of the task.

        Returns:
            str: The data for the task.
        """
        # fmt: off
        query = (
            "SELECT task_id, data "
            "FROM `queue_{}` "
            "WHERE task_id = ?"
        ).format(queue_name)
        # fmt: on
        cur = self._run_query(query, [task_id])
        res = cur.fetchone()

        query = "DELETE FROM `queue_{}` WHERE task_id = ?".format(queue_name)
        self._run_query(query, [task_id])

        return res[1]



import datetime
import json
import time
import uuid

from .constants import WAITING, SUCCESS, FAILED, RETRYING, CANCELED
from .exceptions import MultipleDelayError
from .utils import determine_module, determine_name, import_attr


class Task(object):
    def __init__(
        self,
        task_id=None,
        retries=0,
        is_async=True,
        on_start=None,
        on_success=None,
        on_error=None,
        depends_on=None,
        delay_by=None,
        delay_until=None,
    ):
        """
        A base class for managing the execution & serialization of tasks.

        Ex::

            from alligator import Task

            # Create the task itself.
            task = Task()
            # ...or...
            task = Task(task_id='my-unique-id', retries=4, is_async=False)

            # Hook up what will be called when the task comes off the queue.
            task.to_call(email_followers, emails=user.followers.emails())

        Args:
            task_id: (str): Optional. A unique identifier for the task.
                Defaults to `None` (create a `uuid4`).
            retries (int): Optional. The number of times to retry a task if it
                fails (throws an exception). Defaults to `0` (no retries).
            is_async (bool): Optional. If the task should be run
                asynchronously or not. Defaults to `True`.
            on_start (callable): Optional. A hook function to run when the
                task is first pulled off the queue. Defaults to `None`.
            on_success (callable): Optional. A hook function to run when the
                task completes successfully. Defaults to `None`.
            on_error (callable): Optional. A hook function to run when the
                task is fails. If a non-zero number of retries are provided,
                this will fire *each time* the task fails. Defaults to `None`.
            depends_on (list): Optional. A list of task_ids that must be
                complete before this task will fire. Defaults to `None`.
            delay_by (int): Optional. The number of seconds to delay before
                the task can be processed. *Mutually exclusive* with
                `delay_until`.
            delay_until (float|datetime|date): Optional. The Unix timestamp
                (or a UTC datetime/date object) to delay processing the task
                until. *Mutually exclusive* with `delay_by`.
        """
        self.task_id = task_id
        self.retries = int(retries)
        self.is_async = is_async
        self.status = WAITING
        self.on_start = on_start
        self.on_success = on_success
        self.on_error = on_error
        self.depends_on = depends_on
        self.delay_until = delay_until
        self.result = None

        if self.delay_until is not None:
            if isinstance(
                self.delay_until, (datetime.datetime, datetime.date)
            ):
                self.delay_until = time.mktime(self.delay_until.timetuple())

        # If the convenience option `delay_by` is seen, calculate the correct
        # `delay_until`.
        if delay_by is not None:
            # The delay options are exclusive.
            if delay_until is not None:
                raise MultipleDelayError(
                    "Only one of 'delay_by' or 'delay_until' can be used at "
                    "a time."
                )

            # Compute the desired timestamp.
            self.delay_until = time.time() + delay_by

        self.func = None
        self.func_args = []
        self.func_kwargs = {}

        if self.task_id is None:
            self.task_id = str(uuid.uuid4())

    def to_call(self, func, *args, **kwargs):
        """
        Sets the function & its arguments to be called when the task is
        processed.

        Ex::

            task.to_call(my_function, 1, 'c', another=True)

        Args:
            func (callable): The callable with business logic to execute
            args (list): Positional arguments to pass to the callable task
            kwargs (dict): Keyword arguments to pass to the callable task
        """
        self.func = func
        self.func_args = args
        self.func_kwargs = kwargs

    def to_waiting(self):
        """
        Sets the task's status as "waiting".

        Useful for the `on_start/on_success/on_failed` hook methods for
        figuring out what the status of the task is.
        """
        self.status = WAITING

    def to_success(self):
        """
        Sets the task's status as "success".

        Useful for the `on_start/on_success/on_failed` hook methods for
        figuring out what the status of the task is.
        """
        self.status = SUCCESS

    def to_failed(self):
        """
        Sets the task's status as "failed".

        Useful for the `on_start/on_success/on_failed` hook methods for
        figuring out what the status of the task is.
        """
        self.status = FAILED

    def to_canceled(self):
        """
        Sets the task's status as "canceled".

        Useful for the `on_start/on_success/on_failed` hook methods for
        figuring out what the status of the task is.
        """
        self.status = CANCELED

    def to_retrying(self):
        """
        Sets the task's status as "retrying".

        Useful for the `on_start/on_success/on_failed` hook methods for
        figuring out what the status of the task is.
        """
        self.status = RETRYING

    def serialize(self):
        """
        Serializes the `Task` data for storing in the queue.

        All data must be JSON-serializable in order to be stored properly.

        Returns:
            str: A JSON string of the task data.
        """
        data = {
            "task_id": self.task_id,
            "retries": self.retries,
            "is_async": self.is_async,
            "module": determine_module(self.func),
            "callable": determine_name(self.func),
            "args": self.func_args,
            "kwargs": self.func_kwargs,
            "options": {},
        }

        if self.on_start:
            data["options"]["on_start"] = {
                "module": determine_module(self.on_start),
                "callable": determine_name(self.on_start),
            }

        if self.on_success:
            data["options"]["on_success"] = {
                "module": determine_module(self.on_success),
                "callable": determine_name(self.on_success),
            }

        if self.on_error:
            data["options"]["on_error"] = {
                "module": determine_module(self.on_error),
                "callable": determine_name(self.on_error),
            }

        if self.delay_until:
            data["options"]["delay_until"] = self.delay_until

        return json.dumps(data)

    @classmethod
    def deserialize(cls, data):
        """
        Given some data from the queue, deserializes it into a `Task`
        instance.

        The data must be similar in format to what comes from
        `Task.serialize` (a JSON-serialized dictionary). Required keys are
        `task_id`, `retries` & `is_async`.

        Args:
            data (str): A JSON-serialized string of the task data

        Returns:
            Task: A populated task
        """
        data = json.loads(data)
        options = data.get("options", {})

        task = cls(
            task_id=data["task_id"],
            retries=data["retries"],
            is_async=data["is_async"],
        )

        func = import_attr(data["module"], data["callable"])
        task.to_call(func, *data.get("args", []), **data.get("kwargs", {}))

        if options.get("on_start"):
            task.on_start = import_attr(
                options["on_start"]["module"], options["on_start"]["callable"]
            )

        if options.get("on_success"):
            task.on_success = import_attr(
                options["on_success"]["module"],
                options["on_success"]["callable"],
            )

        if options.get("on_error"):
            task.on_error = import_attr(
                options["on_error"]["module"], options["on_error"]["callable"]
            )

        if options.get("delay_until"):
            task.delay_until = options["delay_until"]

        return task

    def run(self):
        """
        Runs the task.

        This fires the `on_start` hook function first (if present), passing
        the task itself.

        Then it runs the target function supplied via `Task.to_call` with
        its arguments & stores the result.

        If the target function succeeded, the `on_success` hook function is
        called, passing both the task & the result to it.

        If the target function failed (threw an exception), the `on_error`
        hook function is called, passing both the task & the exception to it.
        Then the exception is re-raised.

        Finally, it returns itself, with `Task.result` set to the result
        from the target function's execution.

        Returns:
            Task: The processed Task.
        """
        if self.on_start:
            self.on_start(self)

        try:
            self.result = self.func(*self.func_args, **self.func_kwargs)
        except Exception as err:
            self.to_failed()

            if self.on_error:
                self.on_error(self, err)

            raise

        self.to_success()

        if self.on_success:
            self.on_success(self, self.result)

        return self



import math
import time
from urllib.parse import urlparse

import redis


class Client(object):
    def __init__(self, conn_string):
        """
        A Redis-based ``Client``.

        Args:
            conn_string (str): The DSN. The host/port/db are parsed out of it.
                Should be of the format ``redis://host:port/db``
        """
        self.conn_string = conn_string
        bits = urlparse(self.conn_string)
        self.conn = self.get_connection(
            host=bits.hostname,
            port=bits.port,
            db=bits.path.lstrip("/").split("/")[0],
        )

    def get_connection(self, host, port, db):
        """
        Returns a ``StrictRedis`` connection instance.
        """
        return redis.StrictRedis(
            host=host, port=port, db=db, decode_responses=True
        )

    def len(self, queue_name):
        """
        Returns the length of the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                `Gator`` instance.

        Returns:
            int: The length of the queue
        """
        return self.conn.zcard(queue_name)

    def drop_all(self, queue_name):
        """
        Drops all the task in the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.
        """
        task_ids = self.conn.zrange(queue_name, 0, -1)

        for task_id in task_ids:
            self.conn.delete(task_id)

        self.conn.delete(queue_name)

    def push(self, queue_name, task_id, data, delay_until=None):
        """
        Pushes a task onto the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.
            task_id (str): The identifier of the task.
            data (str): The relevant data for the task.
            delay_until (float): Optional. The Unix timestamp to delay
                processing of the task until. Default is `None`.

        Returns:
            str: The task ID.
        """
        if delay_until is None:
            delay_until = math.ceil(time.time())

        self.conn.zadd(queue_name, {task_id: delay_until}, nx=True)
        self.conn.set(task_id, data)
        return task_id

    def pop(self, queue_name):
        """
        Pops a task off the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.

        Returns:
            str: The data for the task.
        """
        now = math.floor(time.time())
        available_to_pop = self.conn.zrangebyscore(
            queue_name, 0, now, start=0, num=1
        )

        if not len(available_to_pop):
            return None

        popped = self.conn.zpopmin(queue_name)
        task_id, delay_until = popped[0][0], popped[0][1]
        data = self.conn.get(task_id)
        self.conn.delete(task_id)
        return data

    def get(self, queue_name, task_id):
        """
        Pops a specific task off the queue by identifier.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                ``Gator`` instance.
            task_id (str): The identifier of the task.

        Returns:
            str: The data for the task.
        """
        self.conn.zrem(queue_name, task_id)
        data = self.conn.get(task_id)

        if data:
            self.conn.delete(task_id)
            return data



import math
import time


class Client(object):
    queues = {}
    task_data = {}

    def __init__(self, conn_string):
        """
        An in-memory `Client`. Useful for development & testing.

        Likely not particularly thread-safe.

        Args:
            conn_string (str): The DSN. Ignored.
        """
        # We ignore the conn_string, since everything is happening in-memory.
        pass

    def len(self, queue_name):
        """
        Returns the length of the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                `Gator` instance.

        Returns:
            int: The length of the queue
        """
        return len(self.__class__.queues.get(queue_name, []))

    def drop_all(self, queue_name):
        """
        Drops all the task in the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                `Gator` instance.
        """
        cls = self.__class__

        for task_id, _ in cls.queues.get(queue_name, []):
            cls.task_data.pop(task_id, None)

        cls.queues[queue_name] = []

    def push(self, queue_name, task_id, data, delay_until=None):
        """
        Pushes a task onto the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                `Gator` instance.
            task_id (str): The identifier of the task.
            data (str): The relevant data for the task.
            delay_until (float): Optional. The Unix timestamp to delay
                execution of the task until. Default is `None` (no delay).

        Returns:
            str: The task's ID
        """
        cls = self.__class__
        cls.queues.setdefault(queue_name, [])
        cls.queues[queue_name].append([task_id, delay_until])
        cls.task_data[task_id] = data
        return task_id

    def pop(self, queue_name):
        """
        Pops a task off the queue.

        Args:
            queue_name (str): The name of the queue. Usually handled by the
                `Gator` instance.

        Returns:
            str: The data for the task.
        """
        cls = self.__class__
        queue = cls.queues.get(queue_name, [])
        now = math.floor(time.time())

        for offset, task_info in enumerate(queue):
            task_id, delay_until = task_info[0], task_info[1]

            # Check for a delay.
            if delay_until is not None:
                if now < delay_until:
                    continue

            # We've found one we can process.
            queue.pop(offset)
            return cls.task_data.pop(task_id, None)

    def get(self, queue_name, task_id):
        """
        Pops a specific task off the queue by identifier.

        Args:
            queue_name: The name of the queue. Usually handled by the
                `Gator` instance.
            task_id (str): The identifier of the task.

        Returns:
            str: The data for the task.
        """
        # This method is *very* non-thread-safe.
        cls = self.__class__
        queue = cls.queues.get(queue_name, [])

        for offset, task_info in enumerate(queue):
            if task_info[0] == task_id:
                queue.pop(offset)
                return cls.task_data.pop(task_id, None)



import logging
import os
import signal
import time
import traceback

from alligator.constants import ALL


class Worker(object):
    def __init__(
        self,
        gator,
        max_tasks=0,
        to_consume=ALL,
        nap_time=0.1,
        log_level=logging.INFO,
    ):
        """
        An object for consuming the queue & running the tasks.

        Ex::

            from alligator import Gator, Worker

            gator = Gator('locmem://')
            worker = Worker(gator)
            worker.run_forever()

        Args:
            gator (Gator): A configured `Gator` object
            max_tasks (int): Optional. The maximum number of tasks to consume.
                Useful if you're concerned about memory leaks or want
                short-lived workers. Defaults to `0` (unlimited tasks).
            to_consume (str): Optional. The queue name the worker should
                consume from. Defaults to `ALL`.
            nap_time (float): Optional. To prevent high CPU usage in the busy
                loop, you can specify a time delay (in seconds) between
                tasks. Set to `0` to disable sleep & consume as fast as
                possible. Defaults to `0.1`
            log_level (int): Optional. The logging level you'd like for
                output. Default is `logging.INFO`.
        """
        self.gator = gator
        self.max_tasks = int(max_tasks)
        self.to_consume = to_consume
        self.nap_time = nap_time
        self.tasks_complete = 0
        self.keep_running = False
        self.log_level = log_level
        self.log = self.get_log(self.log_level)

    def get_log(self, log_level=logging.INFO):
        """
        Sets up logging for the instance.

        Args:
            log_level (int): Optional. The logging level you'd like for
                output. Default is `logging.INFO`.

        Returns:
            logging.Logger: The log instance.
        """
        log = logging.getLogger(__name__)
        default_format = logging.Formatter(
            "%(asctime)s %(name)s %(levelname)s %(message)s"
        )
        stdout_handler = logging.StreamHandler()
        stdout_handler.setFormatter(default_format)
        log.addHandler(stdout_handler)
        log.setLevel(logging.INFO)
        return log

    def ident(self):
        """
        Returns a string identifier for the worker.

        Used in the printed messages & includes the process ID.
        """
        return "Alligator Worker (#{})".format(os.getpid())

    def starting(self):
        """
        Prints a startup message to stdout.
        """
        self.keep_running = True
        ident = self.ident()
        self.log.info(
            '{} starting & consuming "{}".'.format(ident, self.to_consume)
        )

        if self.max_tasks:
            self.log.info(
                "{} will die after {} tasks.".format(ident, self.max_tasks)
            )
        else:
            self.log.info("{} will never die.".format(ident))

    def interrupt(self):
        """
        Prints an interrupt message to stdout.
        """
        self.keep_running = False
        ident = self.ident()
        self.log.info(
            '{} for "{}" saw interrupt. Finishing in-progress task.'.format(
                ident, self.to_consume
            )
        )

    def stopping(self):
        """
        Prints a shutdown message to stdout.
        """
        self.keep_running = False
        ident = self.ident()
        self.log.info(
            '{} for "{}" shutting down. Consumed {} tasks.'.format(
                ident, self.to_consume, self.tasks_complete
            )
        )

    def result(self, result):
        """
        Prints the received result from a task to stdout.

        :param result: The result of the task
        """
        if result is not None:
            self.log.info(result)

    def check_and_run_task(self):
        """
        Handles the logic of checking for & executing a task.

        `Worker.run_forever` uses this in a loop to actually handle the main
        logic, though you can call this on your own if you have different
        needs.

        Returns:
            bool: `True` if a task was run successfully, `False` if there was
                no task to process or executing the task failed.
        """
        if self.gator.len():
            try:
                task = self.gator.pop()
            except Exception as err:
                self.log.exception(err)
                return False

            if task is None:
                return False

            self.tasks_complete += 1
            self.result(task.result)
            return True

        return False

    def run_forever(self):
        """
        Causes the worker to run either forever or until the
        `Worker.max_tasks` are reached.
        """
        self.starting()

        def handle(signum, frame):
            self.interrupt()

        signal.signal(signal.SIGINT, handle)

        while self.keep_running:
            if self.max_tasks and self.tasks_complete >= self.max_tasks:
                self.stopping()
                break

            self.check_and_run_task()

            if self.nap_time >= 0:
                time.sleep(self.nap_time)

        return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alligator', 'alligator.backends']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'alligator',
    'version': '1.0.0b2',
    'description': 'Simple offline task queues.',
    'long_description': 'Alligator\n=========\n\n.. image:: https://travis-ci.org/toastdriven/alligator.png?branch=master\n        :target: https://travis-ci.org/toastdriven/alligator\n\nSimple offline task queues. For Python.\n\n`"See you later, alligator."`\n\nLatest documentation at http://alligator.readthedocs.org/en/latest/.\n\n\nRequirements\n------------\n\n* Python 3.6+\n* (Optional) ``redis`` for the Redis backend\n* (Optional) ``boto3>=1.12.0`` for the SQS backend\n\n\nBasic Usage\n-----------\n\nThis example uses Django, but there\'s nothing Django-specific about Alligator.\n\nI repeat, You can use it with **any** Python code that would benefit from\nbackground processing.\n\n.. code:: python\n\n    from alligator import Gator\n\n    from django.contrib.auth.models import User\n    from django.shortcuts import send_email\n\n\n    # Make a Gator instance.\n    # Under most circumstances, you would configure this in one place &\n    # import that instance instead.\n    gator = Gator(\'redis://localhost:6379/0\')\n\n\n    # The task itself.\n    # Nothing special, just a plain *undecorated* function.\n    def follow_email(followee_username, follower_username):\n        followee = User.objects.get(username=followee_username)\n        follower = User.objects.get(username=follower_username)\n\n        subject = \'You got followed!\'\n        message = \'Hey {}, you just got followed by {}! Whoohoo!\'.format(\n            followee.username,\n            follower.username\n        )\n        send_email(subject, message, \'server@example.com\', [followee.email])\n\n\n    # An simple, previously expensive view.\n    @login_required\n    def follow(request, username):\n        # You\'d import the task function above.\n        if request.method == \'POST\':\n            # Schedule the task.\n            # Use args & kwargs as normal.\n            gator.task(follow_email, request.user.username, username)\n            return redirect(\'...\')\n\n\nRunning Tasks\n-------------\n\nRather than trying to do autodiscovery, fanout, etc., you control how your\nworkers are configured & what they consume.\n\nIf your needs are simple, run the included ``latergator.py`` worker:\n\n.. code:: bash\n\n    $ python latergator.py redis://localhost:6379/0\n\nIf you have more complex needs, you can create a new executable file\n(bin script, management command, whatever) & drop in the following code.\n\n.. code:: python\n\n    from alligator import Gator, Worker\n\n    # Bonus points if you import that one pre-configured ``Gator`` instead.\n    gator = Gator(\'redis://localhost:6379/0\')\n\n    # Consume & handle all tasks.\n    worker = Worker(gator)\n    worker.run_forever()\n\n\nLicense\n-------\n\nNew BSD\n\n\nRunning Tests\n-------------\n\nAlligator has 95%+ test coverage & aims to be passing/stable at all times.\n\nIf you\'d like to run the tests, clone the repo, then run::\n\n    $ virtualenv -p python3 env\n    $ . env/bin/activate\n    $ pip install -r requirements-tests.txt\n    $ python setup.py develop\n    $ pytest -s -v --cov=alligator --cov-report=html tests\n\nThe full test suite can be run via:\n\n    $ export ALLIGATOR_TESTS_INCLUDE_SQS=true\n    $ ./tests/run_all.sh\n\nThis requires all backends/queues to be running, as well as valid AWS\ncredentials if ``ALLIGATOR_TESTS_INCLUDE_SQS=true`` is set.\n\n\nWHY?!!1!\n--------\n\n* Because I have NIH-syndrome.\n* Or because I longed for something simple (~375 loc).\n* Or because I wanted something with tests (90%+ coverage) & docs.\n* Or because I wanted pluggable backends.\n* Or because testing some other queuing system was a pain.\n* Or because I\'m an idiot.\n\n\nRoadmap\n-------\n\n`1.0.0`:\n\n    * Stable release of all the alpha/beta code\n    * Committing to backward-compatibility in the 1.X series\n\nPost-`1.0.0`:\n\n    * Expand the supported backends\n        * Kafka?\n        * ActiveMQ support?\n        * RabbitMQ support?\n        * ???\n',
    'author': 'Daniel Lindsley',
    'author_email': 'daniel@toastdriven.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/toastdriven/alligator',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import importlib

from .exceptions import UnknownModuleError, UnknownCallableError


def determine_module(func):
    """
    Given a function, returns the Python dotted path of the module it comes
    from.

    Ex::

        from random import choice
        determine_module(choice) # Returns 'random'

    Args:
        func (callable): The callable

    Returns:
        str: Dotted path string
    """
    return func.__module__


def determine_name(func):
    """
    Given a function, returns the name of the function.

    Ex::

        from random import choice
        determine_name(choice) # Returns 'choice'

    Args:
        func (callable): The callable

    Returns:
        str: Name string
    """
    if hasattr(func, "__name__"):
        return func.__name__
    elif hasattr(func, "__class__"):
        return func.__class__.__name__

    # This shouldn't be possible, but blow up if so.
    raise AttributeError("Provided callable '{}' has no name.".format(func))


def import_module(module_name):
    """
    Given a dotted Python path, imports & returns the module.

    If not found, raises ``UnknownModuleError``.

    Ex::

        mod = import_module('random')

    Args:
        module_name (str): The dotted Python path

    Returns:
        module: The imported module
    """
    try:
        return importlib.import_module(module_name)
    except ImportError as err:
        raise UnknownModuleError(str(err))


def import_attr(module_name, attr_name):
    """
    Given a dotted Python path & an attribute name, imports the module &
    returns the attribute.

    If not found, raises ``UnknownCallableError``.

    Ex::

        choice = import_attr('random', 'choice')

    Args:
        module_name (str): The dotted Python path
        attr_name (str): The attribute name

    Returns:
        attribute
    """
    module = import_module(module_name)

    try:
        return getattr(module, attr_name)
    except AttributeError as err:
        raise UnknownCallableError(str(err))
